@@ -5,7 +5,7 @@ use Carp;
 use Data::Dump qw( dump );
 use base qw( Search::Tools::Object );
 
-our $VERSION = '0.80';
+our $VERSION = '0.81';
 
 # debugging only
 my $OPEN  = '[';
@@ -160,12 +160,25 @@ sub _as_sentences {
     # this regex is a sanity check for phrases. we replace the \ with a
     # more promiscuous check because the single space is too naive
     # for real text (e.g. st. john's)
-    my $qre = $self->{_qre};
+    my $qre           = $self->{_qre};
+    my @n_terms       = split( m/[\|\ ]/, $qre );
+    my @phrases       = split( m/\|/, $qre );
+    my $min_proximate = 0;
+    for my $p (@phrases) {
+        my $len = $p =~ s/\\ /\\ /g;
+        next unless $len;
+        $min_proximate += ( $len + 1 );    # +1 because terms = spaces+1
+    }
+    my $n_terms = scalar @n_terms;
     my $query_has_phrase = $qre =~ s/(\\ )+/.+/g;
 
-    if ( $self->debug ) {
+    if ($debug) {
         warn "heat_sentence_starts: " . dump($heat_sentence_starts);
         warn "token_list_heat: " . dump($token_list_heat);
+        warn "n_terms: $n_terms";
+        warn "phrases: " . dump( \@phrases );
+        warn "query_has_phrase: $query_has_phrase";
+        warn "min_proximate: $min_proximate";
     }
 
     # find the "sentence" that each hot token appears in.
@@ -181,7 +194,7 @@ sub _as_sentences {
         # a little optimization for when we've got
         # multiple hot tokens in the same sentence
         if ( exists $heat_sentence_ends{$start} ) {
-            $self->debug
+            $debug
                 and warn "found cached end $heat_sentence_ends{$start} "
                 . "for start $start token $token_pos\n";
 
@@ -212,12 +225,12 @@ sub _as_sentences {
         while ( $end < $max_end ) {
             my $tok = $tokens->get_token( $end++ );
             if ( !$tok ) {
-                $self->debug and warn "No token at end=$end";
+                $debug and warn "No token at end=$end";
                 last;
             }
             if ( $tok->is_sentence_end ) {
                 $end--;    # move back one position
-                if ( $self->debug ) {
+                if ($debug) {
                     warn "tok $_ is_sentence_end end=$end";
                     $tok->dump;
                 }
@@ -228,7 +241,7 @@ sub _as_sentences {
         # back up if we've exceeded the 0-based tokens array.
         $end = $num_tokens if $end > $num_tokens;
 
-        $self->debug
+        $debug
             and warn "start=$start max_end=$max_end "
             . "sentence_length=$sentence_length end=$end "
             . "token_pos=$token_pos\n";
@@ -236,7 +249,7 @@ sub _as_sentences {
         # if we didn't yet set the actual hot token,
         # include everything up to it.
         if ( $end < $token_pos ) {
-            $self->debug
+            $debug
                 and warn "resetting end=$token_pos\n";
 
             $end = $token_pos;
@@ -247,7 +260,7 @@ sub _as_sentences {
         $heat_sentence_ends{$start} = $end;
     }
 
-    $self->debug and warn "starts_ends: " . dump( \@starts_ends );
+    $debug and warn "starts_ends: " . dump( \@starts_ends );
 
     my @spans;
     my %seen_pos;
@@ -261,25 +274,25 @@ START_END:
 
         my ( $start, $hot_pos, $end ) = @$start_end;
     POS: for my $pos ( $start .. $end ) {
-            next if $seen_pos{$pos}++;
+            next POS if $seen_pos{$pos}++;
             $heat += ( exists $heatmap{$pos} ? $heatmap{$pos} : 0 );
             push( @cluster_tokens, $tokens->get_token($pos) );
         }
 
         # if we had already seen_pos all positions.
-        next unless @cluster_tokens;
+        next START_END unless @cluster_tokens;
 
         # sanity: make sure we still have something hot
         my $has_hot = 0;
         my @cluster_pos;
         my @strings;
-        for (@cluster_tokens) {
+    TOK: for (@cluster_tokens) {
             my $pos = $_->pos;
             $has_hot++ if exists $heatmap{$pos};
             push @strings,     $_->str;
             push @cluster_pos, $pos;
         }
-        next unless $has_hot;
+        next START_END unless $has_hot;
 
         # the final string is a sentence end,
         # but we only want the first char in it,
@@ -293,19 +306,64 @@ START_END:
         $span{tokens}    = \@cluster_tokens;
         $span{str}       = join( '', @strings );
 
+        # spans with more *unique* hot tokens in a single span rank higher
+        # spans with more *proximate* hot tokens in a single span rank higher
+        my %uniq          = ();
+        my $i             = 0;
+        my $num_proximate = 0;
+        for (@cluster_pos) {
+            if ( exists $heatmap{$_} ) {
+                $uniq{ lc $strings[$i] } += $heatmap{$_};
+                if ( $i && exists $heatmap{ $cluster_pos[ $i - 2 ] } ) {
+                    $num_proximate++;
+                }
+            }
+            $i++;
+        }
+        $span{unique}    = scalar keys %uniq;
+        $span{proximate} = $num_proximate;
+
         # no false phrase matches if !_treat_phrases_as_singles
-        # stemmer check because regex will likely fail when stemmer is on
-        if (    $query_has_phrase
-            and !$self->{_treat_phrases_as_singles}
-            and !$self->{_stemmer} )
+        # stemmer check because regex will likely fail
+        # when stemmer is on
+        if ( $query_has_phrase
+            and !$self->{_treat_phrases_as_singles} )
         {
+            if ( !$self->{_stemmer} ) {
+
+                #warn "_treat_phrases_as_singles NOT true";
+                if ( $span{str} !~ m/$qre/ ) {
+                    $debug
+                        and warn
+                        "treat_phrases_as_singles=FALSE and '$span{str}' failed to match $qre\n";
+                    next START_END;
+                }
+            }
+            else {
+
+                # if stemmer was on, we cannot rely on the regex,
+                # but we assume that number of uniq terms must match query
+
+                if ( $n_terms > $span{unique} ) {
+
+                    $debug
+                        and warn
+                        "treat_phrases_as_singles=FALSE and '$span{str}' "
+                        . "expected $n_terms unique terms, got $span{unique}\n";
+                    next START_END;
+                }
+
+                # edge case: lots of uniques (as for common stem)
+                # but none of them consecutive
+
+                if ( !$num_proximate || $num_proximate < $min_proximate ) {
+                    $debug
+                        and warn
+                        "treat_phrases_as_singles=FALSE and '$span{str}' "
+                        . "num_proximate=$num_proximate, min_proximate=$min_proximate\n";
+                    next START_END;
+                }
 
-            #warn "_treat_phrases_as_singles NOT true";
-            if ( $span{str} !~ m/$qre/ ) {
-                $self->debug
-                    and warn
-                    "treat_phrases_as_singles=FALSE and '$span{str}' failed to match $qre\n";
-                next;
             }
         }
 
@@ -323,17 +381,6 @@ START_END:
             );
         }
 
-        # spans with more *unique* hot tokens in a single span rank higher
-        my %uniq = ();
-        my $i    = 0;
-        for (@cluster_pos) {
-            if ( exists $heatmap{$_} ) {
-                $uniq{ lc $strings[$i] } += $heatmap{$_};
-            }
-            $i++;
-        }
-        $span{unique} = scalar keys %uniq;
-
         push @spans, \%span;
 
     }
@@ -348,11 +395,13 @@ sub _sort_spans {
     return [
 
         # sort by unique,
+        # then by proximity
         # then by heat
         # then by pos
 
         sort {
                    $b->{unique} <=> $a->{unique}
+                || $b->{proximate} <=> $a->{proximate}
                 || $b->{heat} <=> $a->{heat}
                 || $a->{pos}->[0] <=> $b->{pos}->[0]
             } @{ $_[1] }
@@ -365,17 +414,35 @@ sub _no_sentences {
     my $lhs_window = int( $window / 2 );
     my $debug = $self->debug || 0;
 
-    # this regex is a sanity check for phrases. we replace the \ with a
-    # more promiscuous check because the single space is too naive
-    # for real text (e.g. st. john's)
-    my $qre = $self->{_qre};
-    my $query_has_phrase = $qre =~ s/(\\ )+/.+/g;
-
-    # build heatmap
     my $num_tokens      = $tokens->len;
     my $tokens_arr      = $tokens->as_array;
     my %heatmap         = ();
     my $token_list_heat = $tokens->get_heat;
+
+    # this regex is a sanity check for phrases. we replace the \ with a
+    # more promiscuous check because the single space is too naive
+    # for real text (e.g. st. john's)
+    my $qre           = $self->{_qre};
+    my @n_terms       = split( m/[\|\ ]/, $qre );
+    my @phrases       = split( m/\|/, $qre );
+    my $min_proximate = 0;
+    for my $p (@phrases) {
+        my $len = $p =~ s/\\ /\\ /g;
+        next unless $len;
+        $min_proximate += ( $len + 1 );    # +1 because terms = spaces+1
+    }
+    my $n_terms = scalar @n_terms;
+    my $query_has_phrase = $qre =~ s/(\\ )+/.+/g;
+
+    if ($debug) {
+        warn "token_list_heat: " . dump($token_list_heat);
+        warn "n_terms: $n_terms";
+        warn "phrases: " . dump( \@phrases );
+        warn "query_has_phrase: $query_has_phrase";
+        warn "min_proximate: $min_proximate";
+    }
+
+    # build heatmap
     for (@$token_list_heat) {
         my $token = $tokens->get_token($_);
         $heatmap{ $token->pos } = $token->is_hot;
@@ -408,7 +475,8 @@ sub _no_sentences {
         $i++;
     }
 
-    $debug and warn "proximity: $proximity   clusters: " . dump \@clusters;
+    $debug
+        and warn "proximity: $proximity   clusters: " . dump \@clusters;
 
     # create spans from each cluster, each with a weight.
     # we do the initial sort so that clusters that overlap
@@ -447,7 +515,7 @@ CLUSTER:
             pop @cluster_tokens;
         }
 
-        next unless @cluster_tokens;
+        next CLUSTER unless @cluster_tokens;
 
         # sanity: make sure we still have something hot
         my $has_hot = 0;
@@ -459,7 +527,7 @@ CLUSTER:
             push @strings,     $_->str;
             push @cluster_pos, $pos;
         }
-        next unless $has_hot;
+        next CLUSTER unless $has_hot;
 
         $span{cluster} = $cluster;
         $span{heat}    = $heat;
@@ -467,19 +535,60 @@ CLUSTER:
         $span{tokens}  = \@cluster_tokens;
         $span{str}     = join( '', @strings );
 
+        # spans with more *unique* hot tokens in a single span rank higher
+        # spans with more *proximate* hot tokens in a single span rank higher
+        my %uniq          = ();
+        my $i             = 0;
+        my $num_proximate = 0;
+        for (@cluster_pos) {
+            if ( exists $heatmap{$_} ) {
+                $uniq{ lc $strings[$i] } += $heatmap{$_};
+                if ( $i && exists $heatmap{ $cluster_pos[ $i - 2 ] } ) {
+                    $num_proximate++;
+                }
+            }
+            $i++;
+        }
+        $span{unique}    = scalar keys %uniq;
+        $span{proximate} = $num_proximate;
+
         # no false phrase matches if !_treat_phrases_as_singles
         # stemmer check because regex will likely fail when stemmer is on
-        if (    $query_has_phrase
-            and !$self->{_treat_phrases_as_singles}
-            and !$self->{_stemmer} )
+        if ( $query_has_phrase
+            and !$self->{_treat_phrases_as_singles} )
         {
+            if ( !$self->{_stemmer} ) {
+
+                #warn "_treat_phrases_as_singles NOT true";
+                if ( $span{str} !~ m/$qre/ ) {
+                    $debug
+                        and warn
+                        "treat_phrases_as_singles=FALSE and '$span{str}' failed to match $qre\n";
+                    next CLUSTER;
+                }
+            }
+            else {
+
+                # stemmer used, so check unique term count against n_terms
+                if ( $n_terms > $span{unique} ) {
+                    $debug
+                        and warn
+                        "treat_phrases_as_singles=FALSE and '$span{str}' "
+                        . "expected $n_terms but got $span{unique}\n";
+                    next CLUSTER;
+                }
+
+                # edge case: lots of uniques (as for common stem)
+                # but none of them consecutive
+
+                if ( !$num_proximate || $num_proximate < $min_proximate ) {
+                    $debug
+                        and warn
+                        "treat_phrases_as_singles=FALSE and '$span{str}' "
+                        . "num_proximate=$num_proximate, min_proximate=$min_proximate\n";
+                    next CLUSTER;
+                }
 
-            #warn "_treat_phrases_as_singles NOT true";
-            if ( $span{str} !~ m/$qre/ ) {
-                $self->debug
-                    and warn
-                    "treat_phrases_as_singles=FALSE and '$span{str}' failed to match $qre\n";
-                next;
             }
         }
 
@@ -497,17 +606,6 @@ CLUSTER:
             );
         }
 
-        # spans with more *unique* hot tokens in a single span rank higher
-        my %uniq = ();
-        my $i    = 0;
-        for (@cluster_pos) {
-            if ( exists $heatmap{$_} ) {
-                $uniq{ $strings[$i] } += $heatmap{$_};
-            }
-            $i++;
-        }
-        $span{unique} = scalar keys %uniq;
-
         push @spans, \%span;
 
     }
